/*
 * Copyright (C) 2017 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License
 */

package com.android.dialer.calllog;

import android.content.Context;
import com.android.dialer.calllog.datasources.CallLogDataSource;
import com.android.dialer.calllog.datasources.DataSources;
import com.android.dialer.common.LogUtil;
import com.android.dialer.configprovider.ConfigProviderBindings;
import com.android.dialer.inject.ApplicationContext;
import com.google.common.util.concurrent.Futures;
import com.google.common.util.concurrent.ListenableFuture;
import com.google.common.util.concurrent.MoreExecutors;
import java.util.ArrayList;
import java.util.List;
import javax.inject.Inject;
import javax.inject.Singleton;

/**
 * Coordinates work across {@link DataSources}.
 *
 * <p>All methods should be called on the main thread.
 */
@Singleton
public final class CallLogFramework {

  private final Context appContext;
  private final DataSources dataSources;

  @Inject
  CallLogFramework(@ApplicationContext Context appContext, DataSources dataSources) {
    this.appContext = appContext;
    this.dataSources = dataSources;
  }

  /** Performs necessary setup work when the application is created. */
  public void onApplicationCreate() {
    registerContentObservers();
    CallLogConfig.schedulePollingJob(appContext);
  }

  /** Registers the content observers for all data sources. */
  public void registerContentObservers() {
    LogUtil.enterBlock("CallLogFramework.registerContentObservers");

    // This is the same condition used in MainImpl#isNewUiEnabled. It means that bugfood/debug
    // users will have "new call log" content observers firing. These observers usually do simple
    // things like writing shared preferences.
    // TODO(zachh): Find a way to access Main#isNewUiEnabled without creating a circular dependency.
    if (ConfigProviderBindings.get(appContext).getBoolean("is_nui_shortcut_enabled", false)) {
      for (CallLogDataSource dataSource : dataSources.getDataSourcesIncludingSystemCallLog()) {
        dataSource.registerContentObservers();
      }
    } else {
      LogUtil.i("CallLogFramework.registerContentObservers", "not registering content observers");
    }
  }

  /** Disables the framework. */
  public ListenableFuture<Void> disable() {
    LogUtil.enterBlock("CallLogFramework.disable");

    if (!ConfigProviderBindings.get(appContext).getBoolean("is_nui_shortcut_enabled", false)) {
      LogUtil.i("CallLogFramework.disable", "not disabling");
      return Futures.immediateFuture(null);
    }

    for (CallLogDataSource dataSource : dataSources.getDataSourcesIncludingSystemCallLog()) {
      dataSource.unregisterContentObservers();
    }

    // Clear data only after all content observers have been disabled.
    List<ListenableFuture<Void>> allFutures = new ArrayList<>();
    for (CallLogDataSource dataSource : dataSources.getDataSourcesIncludingSystemCallLog()) {
      allFutures.add(dataSource.clearData());
    }
    return Futures.transform(
        Futures.allAsList(allFutures), unused -> null, MoreExecutors.directExecutor());
  }
}
